import { DOCUMENT } from '@angular/common';
import { ChangeDetectionStrategy, Component, ElementRef, EventEmitter, HostBinding, Inject, Input, NgZone, Output, Renderer2, Self, ViewChild, } from '@angular/core';
import { asyncScheduler } from 'rxjs';
import { observeOn, takeUntil } from 'rxjs/operators';
import { DestroyedService, LoopService } from './services';
import { LoaderService } from './services/loader.service';
import { AnimationStore, CanvasStore, EventsStore, InstancesStore, } from './stores';
export class CanvasComponent {
    constructor(canvasStore, animationStore, eventsStore, loopService, ngZone, hostElement, renderer, document, destroyed) {
        this.canvasStore = canvasStore;
        this.animationStore = animationStore;
        this.eventsStore = eventsStore;
        this.loopService = loopService;
        this.ngZone = ngZone;
        this.hostElement = hostElement;
        this.renderer = renderer;
        this.document = document;
        this.destroyed = destroyed;
        this.hostClass = true;
        this.raycaster = {};
        this.created = new EventEmitter();
        canvasStore.setSize({
            width: hostElement.nativeElement.clientWidth,
            height: hostElement.nativeElement.clientHeight,
        });
    }
    set orthographic(v) {
        this.canvasStore.setIsOrthographic(v);
    }
    set linear(v) {
        this.canvasStore.setIsLinear(v);
    }
    set shadows(v) {
        this.canvasStore.setShadows(v);
    }
    ngOnInit() {
        this.ngZone.runOutsideAngular(() => {
            this.canvasStore.initRendererEffect(this.rendererCanvas.nativeElement);
            this.canvasStore.initSceneEffect(this.scene);
            this.canvasStore.initCameraEffect(this.camera);
            this.canvasStore.initRaycasterEffect(this.raycaster);
            this.initWindowResizeListener();
            this.initActiveListener();
        });
    }
    ngOnDestroy() {
        if (this.windowResizeListener) {
            this.windowResizeListener();
        }
    }
    initWindowResizeListener() {
        this.windowResizeListener = this.renderer.listen(this.document.defaultView, 'resize', () => {
            var _a;
            this.canvasStore.windowResizeEffect({
                size: {
                    width: this.hostElement.nativeElement.clientWidth,
                    height: this.hostElement.nativeElement.clientHeight,
                },
                dpr: ((_a = this.document.defaultView) === null || _a === void 0 ? void 0 : _a.devicePixelRatio) || 0,
            });
        });
    }
    initActiveListener() {
        this.canvasStore.active$
            .pipe(takeUntil(this.destroyed), observeOn(asyncScheduler))
            .subscribe((active) => {
            if (active) {
                const { renderer, camera, scene, } = this.canvasStore.getImperativeState();
                if (renderer && camera && scene) {
                    this.created.emit({ gl: renderer, camera, scene });
                    this.eventsStore.connectEffect(renderer.domElement);
                    this.loopService.start();
                }
            }
        });
    }
}
CanvasComponent.decorators = [
    { type: Component, args: [{
                selector: 'ngt-canvas',
                exportAs: 'ngtCanvas',
                template: ` <canvas #rendererCanvas></canvas> `,
                changeDetection: ChangeDetectionStrategy.OnPush,
                providers: [
                    CanvasStore,
                    EventsStore,
                    InstancesStore,
                    AnimationStore,
                    LoopService,
                    DestroyedService,
                    LoaderService,
                ],
                styles: [`
      :host {
        display: block;
        position: relative;
        width: 100%;
        height: 100%;
        overflow: hidden;
      }

      :host canvas {
        display: block;
      }
    `]
            },] }
];
CanvasComponent.ctorParameters = () => [
    { type: CanvasStore, decorators: [{ type: Self }] },
    { type: AnimationStore, decorators: [{ type: Self }] },
    { type: EventsStore, decorators: [{ type: Self }] },
    { type: LoopService, decorators: [{ type: Self }] },
    { type: NgZone },
    { type: ElementRef },
    { type: Renderer2 },
    { type: undefined, decorators: [{ type: Inject, args: [DOCUMENT,] }] },
    { type: DestroyedService }
];
CanvasComponent.propDecorators = {
    hostClass: [{ type: HostBinding, args: ['class.ngt-canvas',] }],
    orthographic: [{ type: Input }],
    linear: [{ type: Input }],
    shadows: [{ type: Input }],
    camera: [{ type: Input }],
    scene: [{ type: Input }],
    raycaster: [{ type: Input }],
    created: [{ type: Output }],
    rendererCanvas: [{ type: ViewChild, args: ['rendererCanvas', { static: true },] }]
};
//# sourceMappingURL=canvas.component.js.map